(*
    NkV - Graphic Viewer
    Copyright (C) 1998-2003 Nobuyuki Honda (Ferdia) <ferdia@imasy.or.jp>

    Compiler: Delphi 3.01a
*)

// todo: top down support
//       A[JCu\location

program NkV;

uses
  Windows, Messages, ShellAPI,
  DIBLib, SPIUtil,
  SUtil, SForm, SimpleList, ConfFile, KeyTable;

{$R *.RES}
{$R nkvres.res}
{$I nkv.inc}

const
  AppName = 'NkV';
  Version = '0.3.1 (2003-04-18)';
  Copyright = 'Copyright (C) 1999-2003 Nobuyuki Honda (Ferdia)'#13#10+
    #13#10+
    'This software is free software.'#13#10+
    'Freely redistributable. Absolutely no warranty.'#13#10;

  IDM_FIT           = 1;
  IDM_ZOOMIN        = 2;
  IDM_ZOOMOUT       = 3;
  IDM_INFO          = 4;
  IDM_WALLP         = 5;
  IDM_WALLPTILE     = 6;
  IDM_STAYONTOP     = 7;
  IDM_ABOUT         = 8;
  IDM_FULLSCREEN    = 9;
  IDM_STRETCH       = 10;
  IDM_READPREV      = 20;
  IDM_READNEXT      = 21;
  IDM_SMOOTHZOOMIN  = 22;
  IDM_SMOOTHZOOMOUT = 23;
  IDM_PLUGINLIST    = 24;
  IDM_HIDEBORDER    = 25;
  IDM_RELOADFILELIST= 26;
  IDM_READPREVFILE  = 27;
  IDM_READNEXTFILE  = 28;
  IDM_SCROLLTOP     = 29;
  IDM_SCROLLBOTTOM  = 30;
  IDM_SCROLLTOPBOTTOM = 31;
  IDM_SCROLLUP      = 32;
  IDM_SCROLLDOWN    = 33;
  IDM_SCROLLLEFT    = 34;
  IDM_SCROLLRIGHT   = 35;
  IDM_ARCHIVEFILELIST = 90;
  IDM_EXIT          = 99;
  IDM_ZOOMSTART     = 1000;
  IDM_ZOOMEND       = 1999;
  IDM_ZOOM25        = 1025;
  IDM_ZOOM50        = 1050;
  IDM_ZOOM100       = 1100;
  IDM_ZOOM150       = 1150;
  IDM_ZOOM200       = 1200;
  IDM_ZOOM300       = 1300;
  IDM_ZOOM400       = 1400;
  IDM_ZOOM800       = 1800;

  Cmds: array[0..30] of TCmd = (
    (Name: 'FitWindow'; Caption: '&Fit Window'; Id: IDM_FIT),
    (Name: 'ZoomIn'; Caption: 'Zoom In (&+)'; Id: IDM_ZOOMIN),
    (Name: 'ZoomOut'; Caption: 'Zoom Out (&-)'; Id: IDM_ZOOMOUT),
    (Name: 'SmoothZoomIn'; Caption: 'Smooth Zoom In'; Id: IDM_SMOOTHZOOMIN),
    (Name: 'SmoothZoomOut'; Caption: 'Smooth Zoom Out'; Id: IDM_SMOOTHZOOMOUT),
    (Name: 'PicInfo'; Caption: 'Picture &Information...'; Id: IDM_INFO),
    (Name: 'FullScreen'; Caption: 'Full &Screen'; Id: IDM_FULLSCREEN),
    (Name: 'StretchMode'; Caption: '&Stretch Mode'; Id: IDM_STRETCH),
    (Name: 'HideBorder'; Caption: 'Hide &Border'; Id: IDM_HIDEBORDER),
    (Name: 'StayOnTop'; Caption: 'Stay &On Top'; Id: IDM_STAYONTOP),
    (Name: 'Zoom50'; Caption: 'Zoom &50%'; Id: IDM_ZOOM50),
    (Name: 'Zoom100'; Caption: 'Zoom &100%'; Id: IDM_ZOOM100),
    (Name: 'Zoom150'; Caption: 'Zoom 150%(&Q)'; Id: IDM_ZOOM150),
    (Name: 'Zoom200'; Caption: 'Zoom &200%'; Id: IDM_ZOOM200),
    (Name: 'Zoom300'; Caption: 'Zoom &300%'; Id: IDM_ZOOM300),
    (Name: 'Zoom400'; Caption: 'Zoom &400%'; Id: IDM_ZOOM400),
    (Name: 'Zoom800'; Caption: 'Zoom &800%'; Id: IDM_ZOOM800),
    (Name: 'NextImage'; Caption: '&Next Image'; Id: IDM_READNEXT),
    (Name: 'PrevImage'; Caption: '&Previous Image'; Id: IDM_READPREV),
    (Name: 'NextFile'; Caption: '&Next File'; Id: IDM_READNEXTFILE),
    (Name: 'PrevFile'; Caption: '&Previous File'; Id: IDM_READPREVFILE),
    (Name: 'ScrollTop'; Caption: 'Scroll Top'; Id: IDM_SCROLLTOP),
    (Name: 'ScrollBottom'; Caption: 'Scroll Bottom'; Id: IDM_SCROLLBOTTOM),
    (Name: 'ScrollTopBottom'; Caption: 'Scroll Top/Bottom'; Id: IDM_SCROLLTOPBOTTOM),
    (Name: 'ScrollUp'; Caption: 'Scroll Up'; Id: IDM_SCROLLUP),
    (Name: 'ScrollDown'; Caption: 'Scroll Down'; Id: IDM_SCROLLDOWN),
    (Name: 'ScrollLeft'; Caption: 'Scroll Left'; Id: IDM_SCROLLLEFT),
    (Name: 'ScrollRight'; Caption: 'Scroll Right'; Id: IDM_SCROLLRIGHT),
    (Name: 'PluginList'; Caption: 'Plugin &List'; Id: IDM_PLUGINLIST),
    (Name: 'About'; Caption: '&About'; Id: IDM_ABOUT),
    (Name: 'Exit'; Caption: 'E&xit'; Id: IDM_EXIT)
  );

  DefaultKeys: array[0..79] of string = (
    'Enter', 'FullScreen',
    'Esc', 'Exit',
    'Space', 'ScrollTopBottom',
    'PgUp', 'ZoomIn',
    'PgDn', 'ZoomOut',
    'End', 'ScrollBottom',
    'Home', 'ScrollTop',
    'Left', 'ScrollLeft',
    'Up', 'ScrollUp',
    'Right', 'ScrollRight',
    'Down', 'ScrollDown',
    'Num1', 'Zoom100',
    'Num2', 'Zoom200',
    'Num3', 'Zoom300',
    'Num4', 'Zoom400',
    'Num5', 'Zoom50',
    'Num8', 'Zoom800',
    '1', 'Zoom100',
    '2', 'Zoom200',
    '3', 'Zoom300',
    '4', 'Zoom400',
    '5', 'Zoom50',
    '8', 'Zoom800',
    'B', 'PrevImage',
    'S-B', 'PrevFile',
    'C-B', 'HideBorder',
    'N', 'NextImage',
    'S-N', 'NextFile',
    'F', 'FitWindow',
    'I', 'PicInfo',
    'P', 'PrevImage',
    'C-P', 'PluginList',
    'Q', 'Zoom150',
    'R', 'ReloadFileList',
    'S', 'StretchMode',
    'T', 'StayOnTop',
    'X', 'ZoomOut',
    'S-X', 'SmoothZoomOut',
    'Z', 'ZoomIn',
    'S-Z', 'SmoothZoomIn'
  );

  MenuTable: array[0..24] of Integer = (
    IDM_READPREV,
    IDM_READNEXT,
    IDM_READPREVFILE,
    IDM_READNEXTFILE,
    -1,
    IDM_FIT,
    IDM_STRETCH,
    -1,
    IDM_ZOOMIN,
    IDM_ZOOMOUT,
    IDM_ZOOM50,
    IDM_ZOOM100,
    IDM_ZOOM150,
    IDM_ZOOM200,
    IDM_ZOOM300,
    IDM_ZOOM400,
    -1,
    IDM_FULLSCREEN,
    IDM_STAYONTOP,
    IDM_HIDEBORDER,
    -1,
    IDM_INFO,
    -1,
    IDM_ABOUT,
    IDM_EXIT
  );

  Scroll = 16; // J[\gp̃XN[
  ZoomTable: array [0..15] of Integer =
    (12, 25, 50, 66, 100, 125, 150, 200, 300, 400, 500, 600, 700, 800, 1200, 1600);

  rfArchive = 99999;

  WallpaperFile = 'wallpaper.bmp';
  PluginListFile = 'pluginlist';
  KeyMapFile = 'keymap';
  ConfigFile = 'config';

  HeadBufSize = 2048; // tH[}bĝ߂ɓǂݍޗ

resourcestring
  err_cantwrite      = 'Cannot write file.';
  err_unknownformat  = 'Unknown format.';
  //err_notfoundinarc  = 'Loadable file not found in archive file.';

type
  PMaxBitmapInfo = ^TMaxBitmapInfo;
  TMaxBitmapInfo = packed record
    bmiHeader: TBitmapInfoHeader;
    case Boolean of
      //True: (bmiColors: array[0..0] of TRGBQuad);
      True: (bmiColors: array[Byte] of TRGBQuad);
      False: (bmiColorsBF: array[0..2] of DWORD);
  end;

  TImg = record
    Loaded:      Boolean;    // 摜[hσtO
    Width:       Integer;    // 摜̕
    Height:      Integer;    // 摜̍
    DispWidth:   Integer;    // \
    DispHeight:  Integer;    // \
    Zoom:        Integer;    // g嗦(%)
    ColorDepth:  Integer;    // Frbg
    FileSize:    Integer;    // t@CTCY
    FileName:    string;     // t@C
    FileCTime:   TFileTime;  // t@C쐬
    FileATime:   TFileTime;  // t@CŏIANZX
    FileMTime:   TFileTime;  // t@CŏIݓ
    MemSize:     Integer;    // TCY
    Format:      string;     // tH[}bg
    SPIHBm:      HLOCAL;     // SPI pnh
    SPIName:     string;     // Susie plug-in 
    SPIFileName: string;     // Susie plug-in t@C
    Memo:        string;     // eLXg
    BitmapInfo:  TMaxBitmapInfo; // BitmapInfo
    pBits:       Pointer; // rbgf[^
  end;

  TMainWindow = class(TSForm)
    SaveWP: TWindowPlacement;
    SaveWS: Integer;
    FFullScreen: Boolean;
    FBorder: Boolean;
    NormalCursor: HCURSOR;
    GrabCursor: HCURSOR;
    oldpt: TPoint; /////
    procedure SetFullScreen(Value: Boolean);
    property FullScreen: Boolean read FFullScreen write SetFullScreen;
    procedure SetBorder(Value: Boolean);
    property Border: Boolean read FBorder write SetBorder;
    procedure ShowAbout;
    procedure SetTitle;
    procedure FitWindow;
    procedure ZoomImage(Value: Integer);
    procedure PopupMenu;
    procedure ScrollImage(x, y: Integer);
    procedure OnPaint;
    procedure OnCommand(Cmd: Integer);
    procedure OnDropFiles(hDrop: HDROP);
    procedure WndProc(var Message: TMessage); override;
  end;

var
  MainWindow: TMainWindow;
  MsgMouseWheel: Integer;
  Left, Top: Integer;
  Image: TImg;
  KeyTbl: TKeyTable;
  Stretch: Boolean = False;
  FileListPos: Integer;
  ArchiveFileListPos: Integer;
  InfoDlg: HWND;
  ArchiveFileName: string;
  AMHandle: THandle;
  FileList: TSimpleList;
  ArcFileList: TSimpleList;
  TmpFileList: TSimpleList;

  KeepViewPositon: Boolean = False;

  Icon: HICON;
  ProfDir: string;

//  SetLayeredWindowAttributes: function(wnd: HWND; crKey: DWORD;
//    bAlpha: BYTE; dwFlag: DWORD): Boolean; stdcall;

function CommaStr(Value: Integer): string;
var
  s: string;
  i: Integer;
begin
  Str(Value, s);
  Result := '';
  for i := 1 to Length(s) do begin
    Result := Result + s[i];
    if ((Length(s) - i) mod 3 = 0) and (Length(s) <> i) then
      Result := Result + ',';
  end;
end;

function IntToKBStr(Value: Integer): string;
var
  u, s: string;
begin
  if Value < 1024 then
  begin
    Result := IntToStr(Value) + ' bytes';
    Exit;
  end;
  if Value >= 1024*1024*1024 then
  begin
    Value := (Value div 1024 div 1024)*100 div 1024;
    u := ' GB';
  end
  else if Value >= 1024*1024 then
  begin
    Value := (Value div 1024)*100 div 1024;
    u := ' MB';
  end
  else if Value >= 1024 then
  begin
    Value := Value*100 div 1024;
    u := ' KB';
  end
  else begin
    Value := Value*100;
    u := ' bytes';
  end;
  Str(Value, s);
  if Value >= 10000 then
    Result := s[1]+s[2]+s[3]
  else if Value >= 1000 then
    Result := s[1]+s[2]+'.'+s[3]
  else if Value >= 100 then
    Result := s[1]+'.'+s[2]+s[3]
  else
    Result := s;
  Result := Result + u;
end;

{
var
  c: Currency;
  u: string;
  a: Integer;
begin
  if Value >= 1024*1024*1024 then begin
    c := Value / 1024 / 1024 / 1024;
    u := ' GB';
  end
  else if Value >= 1024*1024 then begin
    c := Value / 1024 / 1024;
    u := ' MB';
  end
  else if Value >= 1024 then begin
    c := Value / 1024;
    u := ' KB';
  end
  else begin
    c := Value;
    u := ' bytes';
  end;
  if c >= 100 then
    a := 0
  else if c >= 10 then
    a := 1
  else
    a := 2;
  Str(c :0 :a, Result);
  Result := Result + u;
end;
}

function CountBit(Value: Integer): Integer;
begin
  Result := 0;
  repeat
    Inc(Result, Value and 1);
    Value := Value shr 1
  until Value = 0;
end;

function GetLongFileName(const FileName: string): string;
var
  buf: array [0..MAX_PATH] of Char;
  p: PChar;
  h: THandle;
  FindData: TWin32FindData;
begin
  Result := '';
  GetFullPathName(PChar(FileName), SizeOf(buf), buf, p);
  while True do begin
    h := FindFirstFile(buf, FindData);
    if h = INVALID_HANDLE_VALUE then begin
      Result := string(buf) + Result;
      Exit;
    end;
    Windows.FindClose(h);
    if buf[Length(buf)-1] = ':' then begin
      Result := string(buf) + Result;
      Exit;
    end;
    // '.' 狤L(@Win95a)??
    if string(FindData.cFileName) = '.' then begin
      Result := string(buf) + Result;
      Exit;
    end;
    Result := '\' + string(FindData.cFileName) + Result;
    GetFullPathName(buf, SizeOf(buf), buf, p);
    if p <> nil then
      if ((p-1)^ = '\') then (p-1)^ := #0
  end;
end;

// FileTime(UTC) 烍[J̓𓾂
function FileTimeToLocalTimeStr(const FileTime: TFileTime): string;
var
  buf: array[0..1023] of Char;
  st: TSystemTime;
  ft: TFileTime;
begin
  if (FileTime.dwLowDateTime = 0) and (FileTime.dwHighDateTime = 0) then
  begin
    Result := 'Unknown';
    Exit;
  end;
  FileTimeToLocalFileTime(FileTime, ft);
  FileTimeToSystemTime(ft, st);
  GetDateFormat(LOCALE_USER_DEFAULT, DATE_LONGDATE, @st, nil, buf, SizeOf(buf));
  Result := buf;
  GetTimeFormat(LOCALE_USER_DEFAULT, 0, @st, nil, buf, SizeOf(buf));
  Result := Result + ' ' + buf;
end;

procedure TMainWindow.PopupMenu;
  function Checked(flg: Boolean): Integer;
  begin
    if flg then Result := MF_CHECKED else Result := MF_UNCHECKED;
  end;
var
  pt: TPoint;
  Menu, SubMenu: HMENU;
begin
  Menu := CreatePopupMenu;
  SubMenu := CreateMenu;
  KeyTbl.MenuHandle := Menu;
  KeyTbl.GenerateMenu(MenuTable);
  InsertMenu(Menu, IDM_INFO, MF_BYCOMMAND or MF_POPUP, SubMenu, 'Set &Wallpaper');
  AppendMenu(SubMenu, MF_STRING, IDM_WALLP, '&Normal');
  AppendMenu(SubMenu, MF_STRING, IDM_WALLPTILE, '&Tile');
  InsertMenu(Menu, IDM_INFO, MF_BYCOMMAND or MF_SEPARATOR, SubMenu, nil);

  if not Image.Loaded then
  begin
    EnableMenuItem(Menu, IDM_READPREV, MF_GRAYED);
    EnableMenuItem(Menu, IDM_READNEXT, MF_GRAYED);
    EnableMenuItem(Menu, IDM_READPREVFILE, MF_GRAYED);
    EnableMenuItem(Menu, IDM_READNEXTFILE, MF_GRAYED);
    EnableMenuItem(Menu, IDM_HIDEBORDER, MF_GRAYED);
    EnableMenuItem(Menu, IDM_INFO, MF_GRAYED);
    EnableMenuItem(Menu, IDM_STRETCH, MF_GRAYED);
    EnableMenuItem(Menu, IDM_FULLSCREEN, MF_GRAYED);
    EnableMenuItem(Menu, IDM_FIT, MF_GRAYED);
    EnableMenuItem(Menu, IDM_ZOOMIN, MF_GRAYED);
    EnableMenuItem(Menu, IDM_ZOOMOUT, MF_GRAYED);
    EnableMenuItem(Menu, IDM_ZOOM50, MF_GRAYED);
    EnableMenuItem(Menu, IDM_ZOOM100, MF_GRAYED);
    EnableMenuItem(Menu, IDM_ZOOM150, MF_GRAYED);
    EnableMenuItem(Menu, IDM_ZOOM200, MF_GRAYED);
    EnableMenuItem(Menu, IDM_ZOOM300, MF_GRAYED);
    EnableMenuItem(Menu, IDM_ZOOM400, MF_GRAYED);
    EnableMenuItem(Menu, 21, MF_GRAYED or MF_BYPOSITION);
  end;

  if Stretch then
  begin
    EnableMenuItem(Menu, 4, MF_GRAYED or MF_BYPOSITION);
    EnableMenuItem(Menu, IDM_ZOOMIN, MF_GRAYED);
    EnableMenuItem(Menu, IDM_ZOOMOUT, MF_GRAYED);
  end;

  if FullScreen then
    EnableMenuItem(Menu, IDM_FIT, MF_GRAYED);

  CheckMenuItem(Menu, IDM_FULLSCREEN, Checked(FullScreen));
  CheckMenuItem(Menu, IDM_STAYONTOP, Checked(StayOnTop));
  CheckMenuItem(Menu, IDM_STRETCH, Checked(Stretch));
  CheckMenuItem(Menu, IDM_HIDEBORDER, Checked(not Border));

  GetCursorPos(pt);
  TrackPopupMenu(Menu, TPM_LEFTALIGN or TPM_LEFTBUTTON or TPM_RIGHTBUTTON,
    pt.x, pt.y, 0, Handle, nil);
  DestroyMenu(Menu);
end;

//
// window TCYGɂ킹
//
procedure TMainWindow.FitWindow;
var
  rc: TRect;
begin
  if IsZoomed(Handle) then ShowWindow(Handle, SW_RESTORE);
  rc := Rect(0, 0, Image.DispWidth, Image.DispHeight);
  AdjustWindowRect(rc, GetWindowLong(Handle, GWL_STYLE), False);
  SetSize(rc.Right-rc.Left, rc.Bottom-rc.Top);
end;

//
// vOCTďꏊt@Cɏo
// return : IOResult
//
function MakePluginListFile(const FileName: string): Integer;
var
  hf: THandle;
  fd: TWin32FindData;
  F: Text;
  s, path: string;
begin
  path := GetPluginPath;
  // ꏊ̎擾Ɏs exe ƓfBNgT
  if path = '' then
    path := ExtractFilePath(ParamStr(0));
  if (path[Length(path)] <> '\') or IsDBCSLeadByte(Byte(path[Length(path)-1])) then
    path := path + '\';

  s := '';
  hf := FindFirstFile(PChar(path + '*.spi'), fd);
  if hf <> INVALID_HANDLE_VALUE then begin
    try
      repeat
        s := s + path + fd.cFileName + #13#10;
      until FindNextFile(hf, fd) <> True;
    finally
      Windows.FindClose(hf);
    end;
  end;

  AssignFile(F, FileName);
  {$I-}
  Rewrite(F);
  {$I+}
  Result := IOResult;
  if Result = 0 then
  begin
    Write(F, s);
    CloseFile(F);
  end;
end;

//
// SPI ̃Xgǂ
// return : IOResult
//
function ReadPluginListFile(const FileName: string): Integer;
var
  F: Text;
  p: PChar;
  s: string;
begin
  AssignFile(F, FileName);
  {$I-}
  Reset(F);
  {$I+}
  Result := IOResult;
  if Result = 0 then
  begin
    while not Eof(F) do
    begin
      Readln(F, s);
      if s <> '' then
      begin
        GetMem(p, Length(s)+1);
        Move(PChar(s)^, p^, Length(s)+1);
        PluginList.Add(p);
      end;
    end;
    CloseFile(F);
  end;
end;

//
// Clean Up Image
//
procedure CleanUpImg(var img: TImg);
begin
  with img do begin
    if SPIHBm = 0 then
      GlobalFree(HGLOBAL(pBits))
    else begin
      LocalUnLock(SPIHBm);
      LocalFree(SPIHBm);
    end;
    pBits := nil;
    SPIHBm := 0;
    Loaded := False;
    Memo := '';
    SPIName := '';
    SPIFileName := '';
  end;
end;

procedure TMainWindow.SetTitle;
var
  s, ss: string;
begin
  with Image do
  begin
    if Stretch then
      s := 'Stretch Mode'
    else
      s := IntToStr(Zoom) + '%';
    if ArchiveFileName = '' then
      ss := ExtractFileName(FileName)
    else
      ss := ExtractFileName(ArchiveFileName) + '/' + FileName;
    SetWindowText(Handle, PChar(ss+' ('+IntToStr(Width)+'x'+IntToStr(Height)+'x'+
      IntToStr(ColorDepth)+') '+s+' - '+AppName));
  end;
end;

// pBuffer: 2KBytes炢͂ق
function GetFormatStr(pBuffer: PChar): string;
var
  pBIH: PBitmapInfoHeader;
begin
  Result := '';
  if Copy(pBuffer, 1, 2) = 'BM' then
  begin
    pBIH := @pBuffer[SizeOf(TBitmapFileHeader)];
    if pBIH.biSize = SizeOf(TBitmapInfoHeader) then
    begin
      Result := 'Windows Bitmap';
      case pBIH.biCompression of
        BI_RGB:  Result := Result + ' - RGB';
        BI_RLE8: Result := Result + ' - RLE8';
        BI_RLE4: Result := Result + ' - RLE4';
        BI_BITFIELDS:
          begin
            Result := Result + ' - BITFIELDS';
            with PMaxBitmapInfo(pBIH)^ do
            begin
              Result := Result +' ('+
                IntToStr(CountBit(bmiColorsBF[0]))+
                IntToStr(CountBit(bmiColorsBF[1]))+
                IntToStr(CountBit(bmiColorsBF[2]))+')';
            end;
          end;
      else
        Result := Result + ' - unknown';
      end;
      if pBIH.biHeight < 0 then
        Result := Result + ' (topdown)';
    end
    else if pBIH.biSize = SizeOf(TBitmapCoreHeader) then
      Result := 'OS/2 Bitmap';
  end
  // GIF
  else if Copy(pBuffer, 1, 3) = 'GIF' then
    Result := Copy(pBuffer, 1, 6)
  // JPEG(JFIF)
  else if (Copy(pBuffer, 1, 2) = #$FF#$D8) and (Copy(PChar(@pBuffer[6]), 1, 4) = 'JFIF') then
    Result := 'JPEG - JFIF'
  // Exif?
  else if (Copy(pBuffer, 1, 2) = #$FF#$D8) and (Copy(PChar(@pBuffer[6]), 1, 4) = 'Exif') then
    Result := 'Exif'
  // PNG
  else if Copy(pBuffer, 1, 8) = #$89'PNG'#$0D#$0A#$1A#$0A then
  begin
    Result := 'PNG';
    if pBuffer[28] = #0 then
      Result := Result + ' - non-interlaced'
    else if pBuffer[28] = #1 then
      Result := Result + ' - interlaced';
    if Ord(pBuffer[25]) and 4 <> 0 then
      Result := Result + ' (with alpha)';
  end;
end;

function ToCRLF(const S: string): string;
var
  CRflg: Boolean;
  i: Integer;
begin
  CRflg := False;
  Result := '';
  for i := 1 to Length(S) do
  begin
    if CRflg and (S[i] <> #10) then
      Result := Result + #10
    else if (not CRflg) and (S[i] = #10) then
      Result := Result + #13;
    Result := Result + S[i];
    CRflg := (S[i] = #13);
  end;
end;

procedure ClearListItems(var List: TSimpleList);
var
  i: Integer;
begin
  if List.Count > 0 then
  begin
    for i := 0 to List.Count - 1 do
      FreeMem(List[i]);
  end;
  List.Clear;
end;

//
// t@Cǂ݂
// return : IOResult or -1 (unknown format)
//
function ReadFile(FileName: string; ImgOnly: Boolean;
  const FileList: PFileList; var hAM: THandle): Integer;
var
  F: File;
  size: Integer;
  ptmp: PChar;
  h: THandle;
  fd: TWin32FindData;
  r: Integer;
  oldCursor: HCURSOR;
  HeadBuf: array[0..HeadBufSize-1] of Char;
  BitmapInfo: TMaxBitmapInfo;
  pSrc, pDst: Pointer;
  DstSize: Integer;
  HBm: HLOCAL;
  ColorDepth: Integer;
  TextInfo: string;
  buf: array[0..MAX_PATH] of Char;
  ArcFileName: string;
begin
  oldCursor := SetCursor(LoadCursor(0, IDC_WAIT));
  ptmp := nil;
  FillChar(HeadBuf, SizeOf(HeadBuf), 0);
  if hAM = 0 then
  begin
    // open O(ANZX^CXVO) Find Data 擾
    h := FindFirstFile(PChar(FileName), fd);
    if h <> INVALID_HANDLE_VALUE then Windows.FindClose(h);
    AssignFile(F, FileName);
    {$I-}
    Reset(F, 1);
    {$I+}
    Result := IOResult;
    if Result <> 0 then
    begin
      SetCursor(oldCursor);
      Exit;
    end;
    BlockRead(F, HeadBuf, SizeOf(HeadBuf), r);
    size := FileSize(F);
    ArcFileName := '';
    FileName := GetLongFileName(FileName);
  end else
  begin
    GetFileFromArchive(FileName, ptmp, size, hAM, FileList);
    if size < SizeOf(HeadBuf) then
      Move(ptmp^, HeadBuf, size)
    else
      Move(ptmp^, HeadBuf, SizeOf(HeadBuf));
    FillChar(fd, SizeOf(fd), 0);
    fd.ftLastWriteTime := FileList.FileMTime;
    ArcFileName := FileName;
    FileName := FileList.FileName;
    Result := 0;
  end;

  case GetFormat(FileName, @HeadBuf, h) of
    GF_UNKNOWN:
      Result := -1;
    GF_IMAGE:
      begin
        if hAM = 0 then
        begin
          GetMem(pSrc, size);
          Seek(F, 0);
          BlockRead(F, pSrc^, size);
          CloseFile(F);
        end else
          pSrc := pTmp;

        ExtractImage(h, pSrc, size, PBitmapInfo(@BitmapInfo), pDst, DstSize,
          ColorDepth, TextInfo, HBm);

        CleanUpImg(Image);
        Image.Loaded := True;
        Image.SPIHBm := HBm;
        Image.FileName := FileName;
        Image.FileSize := size;
        Image.MemSize := DstSize;
        Image.BitmapInfo := BitmapInfo;
        Image.Width := BitmapInfo.bmiHeader.biWidth;
        Image.Height := BitmapInfo.bmiHeader.biHeight;
        //if Image.Height < 0 then Image.Height := -Image.Height;
        if Image.Zoom = 0 then Image.Zoom := 100;
        Image.DispWidth  := Image.Width * Image.Zoom div 100;
        Image.DispHeight := Image.Height * Image.Zoom div 100;
        Image.pBits := pDst;
        Image.ColorDepth := ColorDepth;
        Image.Memo := ToCRLF(TextInfo);
        Image.Format := GetFormatStr(pSrc);
        Image.FileCTime  := fd.ftCreationTime;
        Image.FileMTime  := fd.ftLastWriteTime;
        Image.FileATime  := fd.ftLastAccessTime;
        if h <> 0 then
        begin
          GetModuleFileName(h, buf, SizeOf(buf));
          Image.SPIFileName := buf;
          Image.SPIName := GetPluginName(h);
          FreeLibrary(h);
        end;
        FreeMem(pSrc);

        Left := 0;
        Top := 0;
        ArchiveFileName := ArcFileName;
        with MainWindow do
        begin
          SetTitle;
          if not (Stretch or FullScreen or IsZoomed(Handle)) then
            FitWindow;
          if IsWindow(InfoDlg) then
            SendMessage(InfoDlg, WM_INITDIALOG, 0, 0);
          InvalidateRect(Handle, nil, False);
          UpdateWindow(Handle);
        end;
      end;
    GF_ARCHIVE:
      begin
        hAM := h;
        GetFileListFromArchive(hAM, FileName, TmpFileList);
        Result := rfArchive;
      end;
  end;
  SetCursor(oldCursor);
end;

function FileListCompare(Item1, Item2: Pointer): Integer;
begin
  Result := lstrcmp(PFileList(Item1).FileName, PFileList(Item2).FileName);
end;

//
// t@CfBNg̃t@CXg擾
//
procedure GetFileList(const FileName: string);
var
  h: THandle;
  FindData: TWin32FindData;
  s, path: string;
  pItem: PFileList;
begin
  ClearListItems(FileList);
  path := ExtractFilePath(FileName);
  s := path + '*.*';
  h := FindFirstFile(PChar(s), FindData);
  if h = INVALID_HANDLE_VALUE then
    Exit;
  while FindNextFile(h, FindData) = True do
  begin
    if (FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) = 0 then
    begin
      New(pItem);
      lstrcpyn(pItem.FileName, FindData.cFileName, SizeOf(pItem.FileName));
      pItem.IsImage := True; // t@C͉摜t@CƂ݂Ȃ
      FileList.Add(pItem);
    end;
  end;
  Windows.FindClose(h);
  FileList.Sort(FileListCompare);
end;

function ReadNextImageFileFromArchive(FileName: string; StartPos: Integer;
  var FileList: TSimpleList; hAM: THandle): Integer;
var
  i: Integer;
begin
  for i := StartPos to FileList.Count - 1 do
  begin
    if not PFileList(FileList[i]).IsImage then
      Continue;
    if ReadFile(FileName, True, PFileList(FileList[i]), hAM) = 0 then
    begin
      Result := i;
      Exit;
    end else
      PFileList(FileList[i]).IsImage := False;
  end;
  Result := -1;
end;

function ReadPrevImageFileFromArchive(FileName: string; StartPos: Integer;
  var FileList: TSimpleList; hAM: THandle): Integer;
var
  i: Integer;
begin
  for i := StartPos downto 0 do
  begin
    if not PFileList(FileList[i]).IsImage then
      Continue;
    if ReadFile(FileName, True, PFileList(FileList[i]), hAM) = 0 then
    begin
      Result := i;
      Exit;
    end else
      PFileList(FileList[i]).IsImage := False;
  end;
  Result := -1;
end;

function ReadNewFile(const FileName: string; LastFileInArc: Boolean): Integer;
var
  hAM: THandle;
  i: Integer;
begin
  hAM := 0;
  Result := ReadFile(filename, False, nil, hAM);
  if Result = 0 then
  begin
    if AMHandle <> 0 then
      FreeLibrary(AMHandle);
    AMHandle := 0;
  end
  else if Result = rfArchive then
  begin
    if LastFileInArc then
      Result := ReadPrevImageFileFromArchive(filename, TmpFileList.Count-1, TmpFileList, hAM)
    else
      Result := ReadNextImageFileFromArchive(filename, 0, TmpFileList, hAM);
    if Result <> -1 then
    begin
      if AMHandle <> 0 then
        FreeLibrary(AMHandle);
      AMHandle := hAM;
      ClearListItems(ArcFileList);
      for i := 0 to TmpFileList.Count - 1 do
        ArcFileList.Add(TmpFileList[i]);
      TmpFileList.Clear;
      ArchiveFileListPos := Result;
    end;
  end;
end;

//
// ̃t@Cǂ
//
procedure ReadNextFile(ForceNextFile: Boolean);
var
  path, filename: string;
  i, r: Integer;
  dx, dy, w, h: Integer;
begin
  w := Image.Width;
  h := Image.Height;
  dx := Left;
  dy := Top;

  if AMHandle <> 0 then
  begin
    if not ForceNextFile then
    begin
      r := ReadNextImageFileFromArchive(ArchiveFileName, ArchiveFileListPos+1, ArcFileList, AMHandle);
      if r > -1 then
      begin
        ArchiveFileListPos := r;
        if KeepViewPositon and (Image.Width = w) and (Image.Height = h) then
        begin
          Left := dx;
          Top := dy;
        end;
        InvalidateRect(MainWindow.Handle, nil, False);
        UpdateWindow(MainWindow.Handle);
        Exit;
      end;
    end;
    path := ExtractFilePath(ArchiveFileName);
    filename := ArchiveFileName;
  end else
  begin
    path := ExtractFilePath(Image.FileName);
    filename := Image.FileName;
  end;

  if FileList.Count = 0 then
  begin
    GetFileList(filename);
    for i := 0 to FileList.Count - 1 do
    begin
      if path + PFileList(FileList[i]).FileName = filename then
        FileListPos := i;
    end;
  end;

  for i := FileListPos + 1 to FileList.Count - 1 do
  begin
    if not PFileList(FileList[i]).IsImage then
      Continue;
    r := ReadNewFile(Path + PFileList(FileList[i]).FileName, False);
    if r = -1 then
      PFileList(FileList[i]).IsImage := False
    else begin
      FileListPos := i;
      if KeepViewPositon and (Image.Width = w) and (Image.Height = h) then
      begin
        Left := dx;
        Top := dy;
      end;
      InvalidateRect(MainWindow.Handle, nil, False);
      UpdateWindow(MainWindow.Handle);
      Exit;
    end;
  end;
end;

//
// Õt@Cǂ
//
procedure ReadPrevFile(ForcePrevFile: Boolean);
var
  path, filename: string;
  i, r: Integer;
  dx, dy, w, h: Integer;
begin
  w := Image.Width;
  h := Image.Height;
  dx := Left;
  dy := Top;

  if AMHandle <> 0 then
  begin
    if not ForcePrevFile then
    begin
      r := ReadPrevImageFileFromArchive(ArchiveFileName, ArchiveFileListPos-1, ArcFileList, AMHandle);
      if r > -1 then
      begin
        ArchiveFileListPos := r;
        if KeepViewPositon and (Image.Width = w) and (Image.Height = h) then
        begin
          Left := dx;
          Top := dy;
        end;
        InvalidateRect(MainWindow.Handle, nil, False);
        UpdateWindow(MainWindow.Handle);
        Exit;
      end;
    end;
    path := ExtractFilePath(ArchiveFileName);
    filename := ArchiveFileName;
  end else
  begin
    path := ExtractFilePath(Image.FileName);
    filename := Image.FileName;
  end;

  if FileList.Count = 0 then
  begin
    GetFileList(filename);
    for i := 0 to FileList.Count - 1 do
    begin
      if path + PFileList(FileList[i]).FileName = filename then
        FileListPos := i;
    end;
  end;

  for i := FileListPos - 1 downto 0 do
  begin
    if not PFileList(FileList[i]).IsImage then
      Continue;
    r := ReadNewFile(Path + PFileList(FileList[i]).FileName, not ForcePrevFile);
    if r = -1 then
      PFileList(FileList[i]).IsImage := False
    else begin
      FileListPos := i;
      if KeepViewPositon and (Image.Width = w) and (Image.Height = h) then
      begin
        Left := dx;
        Top := dy;
      end;
      InvalidateRect(MainWindow.Handle, nil, False);
      UpdateWindow(MainWindow.Handle);
      Exit;
    end;
  end;
end;

procedure TMainWindow.OnDropFiles(hDrop: HDROP);
var
  FileName: array[0..MAX_PATH-1] of Char;
begin
  DragQueryFile(hDrop, 0, filename, SizeOf(filename));
  DragFinish(hDrop);
  case ReadNewFile(filename, False) of
     0: begin
          ClearListItems(FileList);
          InvalidateRect(Handle, nil, False);
        end;
    -1: MessageBox(Handle, PChar(string(filename) + ':'#13#10 + err_unknownformat),
          AppName, MB_ICONEXCLAMATION);
  end;
end;

//
// On Paint
//
procedure TMainWindow.OnPaint;
var
  dc: HDC;
  ps: TPaintStruct;
  brush: HBrush;
  rc: TRect;
  r, b, fx, fy: Integer;
  bi: TMaxBitmapInfo;
  p: Pointer;
  //s: string;
  //rr, t: Integer;
  //rgn: HRGN;
  //rc2: TRect;
begin
  dc := BeginPaint(Handle, ps);
  rc := ClientRect;
  brush := GetStockObject(GRAY_BRUSH);

  // [hĂȂ΃O[ɓhĂ܂
  if not Image.Loaded then
  begin
    FillRect(dc, rc, brush);
    EndPaint(Handle, ps);
    Exit;
  end;

  if Stretch then
  begin
    with Image do
    begin
      if Width*rc.Bottom > Height*rc.Right then begin
        DispWidth := rc.Right;
        DispHeight := (Height * DispWidth) div Width;
      end
      else begin
        DispHeight := rc.Bottom;
        DispWidth := (Width * DispHeight) div Height;
      end;
    end;
  end;

  fx := 0;
  fy := 0;
  if FullScreen then
  begin
    fx := (GetSystemMetrics(SM_CXSCREEN)-Image.DispWidth) div 2;
    if fx < 0 then fx := 0;
    fy := (GetSystemMetrics(SM_CYSCREEN)-Image.DispHeight) div 2;
    if fy < 0 then fy := 0;
  end;
  SetViewportOrgEx(dc, fx, fy, nil);

  //SetStretchBltMode(dc, COLORONCOLOR);
  //SetMapMode(dc, MM_TEXT);
  with Image do
  begin
    if rc.Right > DispWidth then r := DispWidth else r := rc.Right;
    if rc.Bottom > DispHeight then b := DispHeight else b := rc.Bottom;
    if (Zoom = 100) and not Stretch then begin
      SetDIBitsToDevice(dc, 0, 0, r, b,
        Left, BitmapInfo.bmiHeader.biHeight-b, Top, BitmapInfo.bmiHeader.biHeight,
        pBits, PBitmapInfo(@BitmapInfo)^, DIB_RGB_COLORS);
    end
    else begin
      {
      GetClipRgn(dc, rgn);
      rr := GetRgnBox(rgn, rc2);
      DeleteObject(rgn);
      if rr = SIMPLEREGION then
        SetWindowText(Wnd, 'simple')
      else if rr = COMPLEXREGION then
        SetWindowText(Wnd, 'not simple');
      }
      (*
      t := GetTickCount;
      ResizeDIB(@Bi, Pointer(p), DispWidth, DispHeight,
        Left+ps.rcPaint.Left, Top+ps.rcPaint.Top,
        ps.rcPaint.Right-ps.rcPaint.Left, ps.rcPaint.Bottom - ps.rcPaint.Top,
        PBitmapInfo(@BitmapInfo)^, pBits);
      t := GetTickCount - t;

      with ps.rcPaint do
        s := IntToStr(Left) + ',' + IntToStr(Top) + '-' +
          IntToStr(Right) + ',' + IntToStr(Bottom) + '('+inttostr(t);
      SetWindowText(Wnd, pchar(s));

      SetDIBitsToDevice(dc, ps.rcPaint.Left, ps.rcPaint.Top, ps.rcPaint.Right-ps.rcPaint.Left, b,
        0, 0, 0, Bi.bmiHeader.biHeight, p, PBitmapInfo(@Bi)^, DIB_RGB_COLORS);
      //SetDIBitsToDevice(dc, 0, 0, r, b,
      //  0, 0, 0, Bi.bmiHeader.biHeight, p, PBitmapInfo(@Bi)^, DIB_RGB_COLORS);
      *)

      ResizeDIB(@Bi, Pointer(p), DispWidth, DispHeight, Left, Top,
        rc.Right, rc.Bottom,
       // r, b,
        PBitmapInfo(@BitmapInfo)^, pBits);
      SetDIBitsToDevice(dc, 0, 0, r, b,
        0, 0, 0, Bi.bmiHeader.biHeight, p, PBitmapInfo(@Bi)^, DIB_RGB_COLORS);

      FreeMem(p);
    end;
  end;

  with Image do begin
    // ܂ʂʂ
    if rc.right > DispWidth then begin
      FillRect(dc, Rect(rc.right, 0, DispWidth, DispHeight), brush);
      FillRect(dc, Rect(-fx, -fy, 0, DispHeight+fy+1), brush);
    end;
    if rc.bottom > DispHeight then begin
      FillRect(dc, Rect(0, rc.bottom, rc.Right, DispHeight), brush);
      FillRect(dc, Rect(0, -fy, rc.Right, 0), brush);
    end;
  end;

  EndPaint(Handle, ps);
end;

//
// Picture Information Dialog Proc.
//
function InfoDlgProc(hwndDlg: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
const
  labels = 'Location:'#13#10#13#10+
           'Size:'#13#10+
           'Colors:'#13#10+
           'Format:'#13#10#13#10+
           'File Size:'#13#10+
           'Created:'#13#10+
           'Modified:'#13#10+
           'Accessed:'#13#10#13#10+
           'Memory: '#13#10+
           'Plug-in:';
var
  s, t, path: string;
  dc: HDC;
  ps: TPaintStruct;
  id: Word;
//  h: THandle;
begin
  Result := LRESULT(True);
  case Msg of
    WM_INITDIALOG:
      with Image do
      begin
        {
        SetWindowLong(hwndDlg, GWL_EXSTYLE, GetWindowLong(hwndDlg, GWL_EXSTYLE) or $80000);
        h := LoadLibrary('user32.dll');
        SetLayeredWindowAttributes := GetProcAddress(h, 'SetLayeredWindowAttributes');
        if Assigned(SetLayeredWindowAttributes) then
          SetLayeredWindowAttributes(hwndDlg, 0, 192, 2);
        FreeLibrary(h);
        }
        SetWindowText(hwndDlg, PChar(AppName+' - Picture Information'));
        SetDlgItemText(hwndDlg, IDC_FILENAME, PChar(ExtractFileName(FileName)));
        if icon <> 0 then DestroyIcon(icon);
        if ArchiveFileName = '' then
        begin
          Path := ExtractFilePath(FileName);
          icon := ExtractAssociatedIcon(HInstance, PChar(FileName), id);
        end else
        begin
          Path := ArchiveFileName + '/' + FileName;
          icon := ExtractAssociatedIcon(HInstance, PChar(ArchiveFileName), id);
        end;
        SetDlgItemText(hwndDlg, IDC_LOCATION, PChar(Path));
        SetDlgItemText(hwndDlg, IDC_INFOLABEL, PChar(labels));
        if ColorDepth < 32 then t := CommaStr(1 shl ColorDepth) else t := '';
        s := #13#10#13#10+
             IntToStr(Width)+'x'+IntToStr(Height)+#13#10+
             t+' ('+IntToStr(ColorDepth)+' bit)'+#13#10+
             Format+#13#10#13#10+
             IntToKBStr(FileSize)+' ('+CommaStr(FileSize)+' bytes)'+#13#10+
             FileTimeToLocalTimeStr(FileCTime)+#13#10+
             FileTimeToLocalTimeStr(FileMTime)+#13#10+
             FileTimeToLocalTimeStr(FileATime)+#13#10#13#10+
             IntToKBStr(MemSize)+' ('+CommaStr(MemSize)+' bytes)'+#13#10+
             SPIName+#13#10+
             SPIFilename;
        SetDlgItemText(hwndDlg, IDC_INFO, PChar(s));
        EnableWindow(GetDlgItem(hwndDlg, IDC_TEXTINFOLABEL), Memo <> '');
        SetDlgItemText(hwndDlg, IDC_TEXTINFO, PChar(Memo));
        InvalidateRect(hwndDlg, nil, True);
        //Result := 0;
      end;
    WM_COMMAND:
      case LOWORD(wParam) of
        IDOK, IDCANCEL: SendMessage(hwndDlg, WM_SYSCOMMAND, SC_CLOSE, 0);
      end;
    WM_PAINT:
      begin
        dc := BeginPaint(hwndDlg, ps);
        DrawIcon(dc, (7*LoWord(GetDialogBaseUnits)*100) div 400,
          (7*HiWord(GetDialogBaseUnits)*100) div 800, icon);
        EndPaint(hwndDlg, ps);
      end;
    WM_CLOSE:
      begin
        if icon <> 0 then DestroyIcon(icon);
        DestroyWindow(hwndDlg);
      end;
  else
    Result := 0;
  end;
end;

procedure TMainWindow.ShowAbout;
var
  mbp: TMsgBoxParams;
begin
  FillChar(mbp, SizeOf(mbp), 0);
  with mbp do begin
    cbSize      := SizeOf(mbp);
    hwndOwner   := Handle;
    hInstance   := MainInstance;
    lpszText    := PChar(AppName+' '+Version+#13#10#13#10+Copyright+#13#10+'Profile Folder: '+ProfDir);
    lpszCaption := 'About '+AppName;
    dwStyle     := MB_USERICON;
    lpszIcon    := PChar(1);
  end;
  MessageBoxIndirect(mbp);
end;

procedure TMainWindow.ZoomImage(Value: Integer);
var
  rc: TRect;
begin
  if (Value < 1) or Stretch then Exit;
  with Image do
  begin
    Left := Left * 100 div Zoom;
    Top  := Top * 100 div Zoom;
    Zoom := Value;
    DispWidth := Width * Zoom div 100;
    DispHeight := Height * Zoom div 100;
    Left := Left * Zoom div 100;
    Top  := Top * Zoom div 100;
    // ʒu
    rc := ClientRect;
    if DispWidth > rc.Right then
    begin
     if Left+rc.Right > DispWidth then Left := DispWidth - rc.Right;
    end else
      Left := 0;
    if DispHeight > rc.Bottom then
    begin
      if Top+rc.Bottom > DispHeight then Top := DispHeight - rc.Bottom;
    end else
       Top := 0;

    SetTitle;
  end;
end;

// Scroll Image
// x, y: ړ
procedure TMainWindow.ScrollImage(x, y: Integer);
var
  rc: TRect;
  oldleft, oldtop: Integer;
//  dc: HDC;  rgn: HRGN;   rrc:TRECT;
begin
  with Image do
  begin
    if not Loaded then exit;
    rc := ClientRect;
    oldleft := Left;
    oldtop := Top;

    if DispWidth > rc.Right then begin
      Inc(Left, x);
      if Left < 0 then Left := 0;
      if (Left > DispWidth - rc.Right) then
         Left := DispWidth - rc.Right;
    end;
    if DispHeight > rc.Bottom then begin
      Inc(Top, y);
      if Top < 0 then Top := 0;
      if Top > DispHeight - rc.Bottom then
        Top := DispHeight - rc.Bottom;
    end;
    if ScrollWindow(Handle, (oldleft-Left), (oldtop-Top), nil, @rc) then
      UpdateWindow(Handle);
    //InvalidateRect(hWnd, nil, False);

{
    dc := GetDC(hWnd);
    ScrollDC(dc, (oldleft-Left), (oldtop-Top), rc, rc, 0, @rrc);
    ReleaseDC(hWnd, dc);
    InvalidateRect(hWnd, @rrc, False);
}
  end;
end;

//
// Set Wallpaper
//
function SetWallPaper(const Img: TImg; TileFlg: Boolean): Integer;
var
  F: File;
  s: string;
  bfh: TBitmapFileHeader;
  bmisize: Integer;
  hkResult: HKEY;
  data: array[0..MAX_PATH] of Char;
begin
  s := ProfDir+WallpaperFile;
  AssignFile(F, s);
  {$I-}
  ReWrite(F, 1);
  {$I+}
  Result := IOResult;
  if Result <> 0 then exit;

  with Img do
  begin
    bmisize := SizeOf(BitmapInfo.bmiHeader);
    if BitmapInfo.bmiHeader.biBitCount <= 8 then
      Inc(bmisize, (1 shl BitmapInfo.bmiHeader.biBitCount) * SizeOf(TRGBQUAD));
    with bfh do
    begin
      bfType := byte('B') + byte('M') shl 8;
      bfSize := SizeOf(bfh)+bmisize+BitmapInfo.bmiHeader.biSizeImage;
      bfReserved1 := 0;
      bfReserved2 := 0;
      bfOffBits  := SizeOf(bfh)+bmisize;
    end;
    BlockWrite(F, bfh, SizeOf(bfh));
    BlockWrite(F, BitmapInfo, bmisize);
    BlockWrite(F, pBits^, BitmapInfo.bmiHeader.biSizeImage);
    //if HBInfo <> 0 then LocalUnLock(HBInfo);
  end;
  CloseFile(F);

  if TileFlg then data := '1' else data := '0';
  if RegOpenKeyEx(HKEY_CURRENT_USER, 'Control Panel\desktop', 0, KEY_WRITE, hkResult) = ERROR_SUCCESS then
  begin
    try
      RegSetValueEx(hkResult, 'TileWallPaper', 0, REG_SZ, @data, 1);
    finally
      RegCloseKey(hkResult);
    end;
  end;

  SystemParametersInfo(SPI_SETDESKWALLPAPER, 0, PChar(s),
    SPIF_UPDATEINIFILE or SPIF_SENDWININICHANGE);
end;

procedure TMainWindow.SetFullScreen(Value: Boolean);
begin
  FFullScreen := Value;
  if FFullScreen then
  begin
    FillChar(SaveWP, SizeOf(SaveWP), 0);
    SaveWP.length := SizeOf(SaveWP);
    GetWindowPlacement(Handle, @SaveWP);
    SaveWS := GetWindowLong(Handle, GWL_STYLE);
    SetWindowLong(Handle, GWL_STYLE, WS_OVERLAPPEDWINDOW
      and not WS_CAPTION and not WS_THICKFRAME and not WS_SYSMENU);

    ShowWindow(Handle, SW_SHOWNA);
    MoveWindow(Handle, 0, 0, GetSystemMetrics(SM_CXSCREEN),
      GetSystemMetrics(SM_CYSCREEN), True);

    //ShowWindow(hWnd, SW_SHOWMAXIMIZED);
  end
  else begin
    SetWindowLong(Handle, GWL_STYLE, SaveWS and not WS_MAXIMIZE);
    ShowWindow(Handle, SW_SHOWNA);
    SetWindowPlacement(Handle, @SaveWP);
  end;
end;

procedure TMainWindow.SetBorder(Value: Boolean);
var
  rc: TRect;
begin
  FBorder := Value;
  if FBorder then
    SetWindowLong(Handle, GWL_STYLE, WS_OVERLAPPEDWINDOW)
  else
    SetWindowLong(Handle, GWL_STYLE, WS_OVERLAPPEDWINDOW
      and not WS_CAPTION and not WS_THICKFRAME and not WS_SYSMENU);
  ShowWindow(Handle, SW_SHOWNA);
  rc := ClientRect;
  AdjustWindowRect(rc, GetWindowLong(Handle, GWL_STYLE), False);
  SetSize(rc.Right-rc.Left, rc.Bottom-rc.Top);
end;

function GetNextZoom(zoom: Integer): Integer;
var
  i: Integer;
begin
  Result := zoom;
  for i := 0 to High(ZoomTable) do
  begin
    if ZoomTable[i] > zoom then
    begin
      Result := ZoomTable[i];
      Exit;
    end;
  end;
end;

function GetPrevZoom(zoom: Integer): Integer;
var
  i: Integer;
begin
  Result := zoom;
  for i := High(ZoomTable) downto 0 do
  begin
    if ZoomTable[i] < zoom then
    begin
      Result := ZoomTable[i];
      Exit;
    end;
  end;
end;

procedure TMainWindow.OnCommand(Cmd: Integer);
var
  i: Integer;
  s, ss: string;
  h: HMODULE;
  rc: TRect;
begin
  if Image.Loaded then
  begin
    case cmd of
      IDM_INFO:
        if not (IsWindow(InfoDlg) and IsWindowVisible(InfoDlg)) then
          InfoDlg := CreateDialogParam(hInstance, 'INFODIALOG', Handle, @InfoDlgProc, 0);
      IDM_FIT:
        if not FullScreen then FitWindow;
      IDM_ZOOMIN:
        begin
          ZoomImage(GetNextZoom(Image.Zoom));
          InvalidateRect(Handle, nil, False);
        end;
      IDM_ZOOMOUT:
        begin
          ZoomImage(GetPrevZoom(Image.Zoom));
          InvalidateRect(Handle, nil, False);
        end;
      IDM_ZOOMSTART..IDM_ZOOMEND:
        begin
          ZoomImage(cmd - IDM_ZOOMSTART);
          InvalidateRect(Handle, nil, False);
        end;
      IDM_FULLSCREEN:
        FullScreen := not FullScreen;
      IDM_WALLP:
        SetWallPaper(Image, False);
      IDM_WALLPTILE:
        SetWallPaper(Image, True);
      IDM_STRETCH:
        begin
          Stretch := not Stretch;
          Left := 0;
          Top := 0;
          SetTitle;
          ZoomImage(Image.Zoom);
          InvalidateRect(Handle, nil, False);
        end;
      IDM_READNEXT:
        ReadNextFile(False);
      IDM_READPREV:
        ReadPrevFile(False);
      IDM_READNEXTFILE:
        ReadNextFile(True);
      IDM_READPREVFILE:
        ReadPrevFile(True);
      IDM_SMOOTHZOOMIN:
        for i := 0 to 4 do begin
          ZoomImage(Image.Zoom+1);
          InvalidateRect(Handle, nil, False);
          UpdateWindow(Handle);
        end;
      IDM_SMOOTHZOOMOUT:
        for i := 0 to 4 do begin
          ZoomImage(Image.Zoom-1);
          InvalidateRect(Handle, nil, False);
          UpdateWindow(Handle);
        end;
      IDM_HIDEBORDER:
        if not FullScreen then
          Border := not Border;
      IDM_RELOADFILELIST:
        ClearListItems(FileList);
      IDM_SCROLLTOP:
        ScrollImage(0, -Image.DispHeight);
      IDM_SCROLLBOTTOM:
        ScrollImage(0, Image.DispHeight);
      IDM_SCROLLTOPBOTTOM:
        begin
          rc := ClientRect;
          if Top = Image.DispHeight - rc.Bottom then
            ScrollImage(0, -Image.DispHeight)
          else
            ScrollImage(0, Image.DispHeight);
        end;
      IDM_SCROLLLEFT:  ScrollImage(-Scroll,0);
      IDM_SCROLLRIGHT: ScrollImage(Scroll, 0);
      IDM_SCROLLUP:    ScrollImage(0,-Scroll);
      IDM_SCROLLDOWN:  ScrollImage(0, Scroll);
      IDM_ARCHIVEFILELIST:
        begin
          for i := 0 to ArcFileList.Count - 1 do
            s := s + PChar(ArcFileList[i]) + #13#10;
          MessageBox(Handle, PChar(s), AppName + '', 0);
        end;
    end;
  end;
  case cmd of
    IDM_STAYONTOP:
      StayOnTop := not StayOnTop;
    IDM_PLUGINLIST:
      begin
        s := IntToStr(PluginList.Count) + ' plugin(s) installed.'#13#10#13#10;
        for i := 0 to PluginList.Count - 1 do
        begin
          h := LoadLibrary(PChar(PluginList[i]));
          if h = 0 then
            ss := '(not found)'
          else begin
            ss := GetPluginName(h);
            FreeLibrary(h);
          end;
          s := s + PChar(PluginList[i]) + ': ' + ss + #13#10;
        end;
        MessageBox(Handle, PChar(s), AppName + ' - Plug-in List', 0);
      end;
    IDM_ABOUT:
      ShowAbout;
    IDM_EXIT:
      Close;
  end;
end;

procedure TMainWindow.WndProc(var Message: TMessage);
var
  pt: TPoint;
  x, y: Integer;
begin
  with Message do
  begin
    Result := 0;
    if (Msg = WM_MOUSEWHEEL) or ((Msg = MsgMouseWheel) and (MsgMouseWheel <> 0)) then
    begin
      if (GetKeyState(VK_LBUTTON) < 0) and not Stretch then
      begin
        if wParam > 0 then
          OnCommand(IDM_SMOOTHZOOMIN)
        else
          OnCommand(IDM_SMOOTHZOOMOUT)
      end
      else begin
        if wParam > 0 then
          OnCommand(IDM_READPREV)
        else
          OnCommand(IDM_READNEXT);
      end;
      Exit;
    end;
    case Msg of
      WM_DROPFILES:
        OnDropFiles(wParam);
      WM_PAINT:
        OnPaint;
      {
      WM_CONTEXTMENU:
        DispMenu(hWnd);
      }
      WM_COMMAND:
        OnCommand(LoWord(wParam));
      WM_NCHITTEST:
        if Border or FullScreen then
          inherited
        else
          Result := HTCAPTION;
      WM_LBUTTONDOWN:
        begin
          SetCapture(Handle);
          SetCursor(GrabCursor);
          GetCursorPos(pt);
          oldpt := pt;
        end;
      WM_LBUTTONUP:
        begin
          ReleaseCapture;
          SetCursor(NormalCursor);
        end;
      WM_LBUTTONDBLCLK:
        if not FullScreen then FitWindow;
      WM_RBUTTONUP:
        PopupMenu;
      WM_NCRBUTTONUP:
        if not Border then PopupMenu;
      WM_NCLBUTTONDBLCLK:
        if (not FullScreen) and (not Border) then
          FitWindow
        else
          inherited;
      WM_MOUSEMOVE:
        begin
          // StretchŃXN[悤ƂĂcaption_uNbN
          // ő剻悤ƂƗ̂Stretch ModêƂ͉Ȃ
          if Image.Loaded and (wParam = MK_LBUTTON) and not Stretch then
          begin
            GetCursorPos(pt); // ƂȂ window ÔƂWƂȂ
            ScrollImage(oldpt.x - pt.x, oldpt.y - pt.y);
            oldpt := pt;
          end;
        end;
      WM_SIZE:
        with Image do begin
          if Loaded and not Stretch then begin
            x := LoWord(lParam);
            y := HiWord(lParam);
            if DispWidth > x then begin
              if Left+x > DispWidth then
                Left := DispWidth - x;
            end else
              Left := 0;
            if DispHeight > y then begin
              if Top+y > DispHeight then
                Top := DispHeight - y;
            end else
              Top := 0;
          end;
        end;
      WM_CLOSE:
        begin
          //if pPluginList <> nil then FreeMem(pPluginList);
          CleanUpImg(Image);
          if AMHandle <> 0 then
            FreeLibrary(AMHandle);
          DestroyWindow(Handle);
        end;
      WM_DESTROY:
        PostQuitMessage(0);
    else
      inherited;
    end;
  end;
end;

procedure ReadConfigFile;
var
  ConfFile: TConfFile;
begin
  ConfFile := TConfFile.Create(ProfDir + ConfigFile);
  while ConfFile.GetLine do
  begin
    with ConfFile.Line do
    begin
      if Name = 'KeepViewPosition' then
        KeepViewPositon := lstrcmpi(Args[0], 'on') = 0;
    end;
  end;
  ConfFile.Free;
end;

var
  Msg: TMsg;
  r: Integer;
  s: string;
begin
  ProfDir := GetProfDir('nkv');
  ReadConfigFile;

  KeyTbl := TKeyTable.Create(Cmds, DefaultKeys, ProfDir + KeyMapFile);

  FileMode := 0; // Reset  read only mode ŊJ
  MsgMouseWheel := RegisterWindowMessage('MSWHEEL_ROLLMSG'); // for Win95

  MainWindow := TMainWindow.Create(AppName);
  with MainWindow do
  begin
    NormalCursor := LoadCursor(HInstance, 'HANDOPEN');
    GrabCursor := LoadCursor(HInstance, 'HANDHOLD');
    Cursor := NormalCursor;
    ClassStyle := CS_HREDRAW or CS_VREDRAW or CS_DBLCLKS;
    ExStyle := WS_EX_ACCEPTFILES;
    Border := True;
    Brush := GetStockObject(NULL_BRUSH);
    CreateWnd(CW_USEDEFAULT, CW_USEDEFAULT, 320, 240);
    if Handle = 0 then
      Halt(255);

    FileList := TSimpleList.Create;
    ArcFileList := TSimpleList.Create;
    TmpFileList := TSimpleList.Create;
    PluginList := TSimpleList.Create;

    s := ProfDir + PluginListFile;
    if not FileExist(s) then
    begin
      if MakePluginListFile(s) <> 0 then
        MessageBox(GetDesktopWindow, PChar(s + ': ' + err_cantwrite),
          AppName, MB_ICONEXCLAMATION);
    end;
    ReadPluginListFile(s);

    if ParamStr(1) <> '' then
    begin
      r := ReadNewFile(GetLongFileName(ParamStr(1)), False);
      if r = -1 then
      begin
        MessageBox(Handle, PChar(GetLongFileName(ParamStr(1)) + ':'#13#10 + err_unknownformat),
          AppName, MB_ICONEXCLAMATION);
        CleanUpImg(Image);
        DestroyWindow(Handle);
      end
    end;

    Show;

    while Integer(GetMessage(Msg, 0, 0, 0)) <> 0 do
    begin
      if (InfoDlg = 0 ) or not IsDialogMessage(InfoDlg, Msg) then
      begin
        if not Boolean(TranslateAccelerator(Handle, KeyTbl.AccHandle, Msg)) then
        begin
          TranslateMessage(Msg);
          DispatchMessage(Msg);
        end;
      end;
    end;

    ClearListItems(PluginList);
    PluginList.Free;
    ClearListItems(FileList);
    FileList.Free;
    ClearListItems(ArcFileList);
    ArcFileList.Free;
    ClearListItems(TmpFileList);
    TmpFileList.Free;
    KeyTbl.Free;
    ExitCode := Msg.wParam;
  end;
end.

